/* $Id: log.c,v 1.1.1.1 2005/05/28 18:21:19 bedlam Exp $ */
/* log.c
 * - Functions for logging in ices
 * Copyright (c) 2000 Alexander Havng
 * Copyright (c) 2001 Brendan Cully
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <stdarg.h>
#include <stdlib.h>
#include "log.h"
#include "xml_config.h"
#include "util.h"

/* Private function declarations */
static void mediabox404_log_string (char *format, char *string);
static int mediabox404_log_open_logfile (void);
static int mediabox404_log_close_logfile (void);

extern mediabox404_config_t mediabox404_config;

#define BUFSIZE 4096
static char lasterror[BUFSIZE];
/* Public function definitions */

/* Initialize the log module, creates log file */
void
mediabox404_log_initialize (void)
{
	if (!mediabox404_log_open_logfile ()) {
		mediabox404_log ("%s", mediabox404_log_get_error ());
	}
	mediabox404_log ("Logfile opened");
}

/* Shutdown the log module, close the logfile */
void
mediabox404_log_shutdown (void)
{
	if (!mediabox404_log_close_logfile ()) {
		mediabox404_log ("%s", mediabox404_log_get_error ());
	}
}

/* Close everything, start up with clean slate when 
 * run as a daemon 
void
mediabox404_log_daemonize (void)
{
	close (0); 
	close (1); 
	close (2);

	freopen ("/dev/null", "r", stdin);
	freopen ("/dev/null", "w", stdout);
	freopen ("/dev/null", "w", stderr);

	mediabox404_log_reopen_logfile ();
}
*/

/* Cycle the logfile, usually called from the SIGHUP handler */
int
mediabox404_log_reopen_logfile (void)
{
	mediabox404_log_close_logfile ();
	return mediabox404_log_open_logfile ();
}

/* Log only if verbose mode is set. Prepend output with DEBUG:  */
void
mediabox404_log_debug (const char *fmt, ...)
{
	char buff[BUFSIZE];
	va_list ap;
	time_t now;
	char tmp[1024];
	char *timestamp;

	if (!mediabox404_config.verbose)
		return;

	va_start(ap, fmt);
#ifdef HAVE_VSNPRINTF
	vsnprintf(buff, BUFSIZE, fmt, ap);
#else
	vsprintf(buff, fmt, ap);
#endif
	va_end(ap);

	(void)time(&now);
	timestamp = ctime(&now) + 4;
	*(timestamp + 15) = '\0';
	snprintf(tmp, 1024, "%s DEBUG", timestamp);
	strncat(tmp, ": %s\n", 1024);
	mediabox404_log_string (tmp, buff);
}

/* Log to console and file */
void
mediabox404_log (const char *fmt, ...)
{
	char buff[BUFSIZE];
	va_list ap;
	time_t now;
	char tmp[1024];
	char *timestamp;

	va_start(ap, fmt);
#ifdef HAVE_VSNPRINTF
	vsnprintf(buff, BUFSIZE, fmt, ap);
#else
	vsprintf(buff, fmt, ap);
#endif
	va_end(ap);

	(void)time(&now);
	timestamp = ctime(&now) + 4;
	*(timestamp + 15) = '\0';
	snprintf(tmp, 1024, "%s", timestamp);
	strncat(tmp, " %s\n", 1024);
	mediabox404_log_string (tmp, buff);
}

/* Store error information in module memory */
void
mediabox404_log_error (const char *fmt, ...)
{
	char buff[BUFSIZE];
	va_list ap;
	time_t now;
	char tmp[BUFSIZE+16];
	char *timestamp;

	va_start(ap, fmt);
#ifdef HAVE_VSNPRINTF
	vsnprintf(buff, BUFSIZE, fmt, ap);
#else
	vsprintf(buff, fmt, ap);
#endif
	va_end(ap);

	(void)time(&now);
	timestamp = ctime(&now) + 4;
	*(timestamp + 15) = '\0';
	snprintf(tmp, 1024, "%s", timestamp);
	strncat(tmp, buff, BUFSIZE);
	strncpy (lasterror, tmp, BUFSIZE+16);
}

/* Get last error from log module */
char *
mediabox404_log_get_error (void)
{
	return lasterror;
}

/* Private function definitions */

/* Function to log string to both console and file */
static void
mediabox404_log_string (char *format, char *string)
{
	if (mediabox404_config.logfile) {
		fprintf (mediabox404_config.logfile, format, string);
#ifndef HAVE_SETLINEBUF
		fflush (mediabox404_config.logfile);
#endif
	}
	
	/* Don't log to console when daemonized 
	   if (!mediabox404_config.daemon) {
	*/
	fprintf (stdout, format, string);
	/*
	  }
	*/
}

/* Open the mediabox404 logfile, create it if needed */
static int
mediabox404_log_open_logfile (void)
{
  char namespace[1024], buf[1024];
  FILE *logfp;

  if (! mediabox404_config.base_directory ||
      strlen (mediabox404_config.base_directory) > 1016)
  {
    mediabox404_log_error ("Base directory is invalid");
    return 0;
  }

  snprintf (namespace, sizeof (namespace), "%s/mediabox404.log",
	    mediabox404_config.base_directory);

  logfp = fopen (namespace, "a");

  if (!logfp) {
    mediabox404_log_error ("Error while opening %s, error: %s", namespace,
		    mediabox404_util_strerror (errno, buf, 1024));
    return 0;
  }

  mediabox404_config.logfile = logfp;
#ifdef HAVE_SETLINEBUF
  setlinebuf (mediabox404_config.logfile);
#endif

  return 1;
}

/* Close mediabox404' logfile */
static int
mediabox404_log_close_logfile (void)
{
	if (mediabox404_config.logfile) {
		mediabox404_util_fclose (mediabox404_config.logfile);
	}

	return 1;
}
