/* $Id: reencode.c,v 1.2 2005/06/25 00:29:59 bedlam Exp $ */
/* reencode.c
 * - Functions for reencoding in ices
 * Copyright (c) 2000 Alexander Havng
 * Copyright (c) 2001-3 Brendan Cully
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "definitions.h"
#include "metadata.h"

#ifdef HAVE_LAME_LAME_H
# include <lame/lame.h>
#else
# include <lame.h>
#endif

#include <time.h>
#include <ogg/ogg.h>
#include <vorbis/codec.h>
#include <vorbis/vorbisenc.h>

extern stream404_config_t stream404_config;

/* Global function definitions */

void vorbis_encode_data_float(vorbis_encoder_t *s, float **pcm, int samples)
{
    float **buf;
    int i;

    buf = vorbis_analysis_buffer(&s->vd, samples); 

    for(i=0; i < s->vi.channels; i++)
    {
        memcpy(buf[i], pcm[i], samples*sizeof(float));
    }

    vorbis_analysis_wrote(&s->vd, samples);

    s->samples_in_current_page += samples;
}

/* Requires little endian data (currently) */
void vorbis_encode_data(vorbis_encoder_t *s, signed char *buf, int bytes, int bigendian)
{
    float **buffer;
    int i,j;
    int channels = s->vi.channels;
    int samples = bytes/(2*channels);

    buffer = vorbis_analysis_buffer(&s->vd, samples);

    if(bigendian)
    {
        for(i=0; i < samples; i++)
        {
            for(j=0; j < channels; j++)
            {
                buffer[j][i]=((buf[2*(i*channels + j)]<<8) |
                              (0x00ff&(int)buf[2*(i*channels + j)+1]))/32768.f;
            }
        }
    }
    else
    {
        for(i=0; i < samples; i++)
        {
            for(j=0; j < channels; j++)
            {
                buffer[j][i]=((buf[2*(i*channels + j) + 1]<<8) |
                              (0x00ff&(int)buf[2*(i*channels + j)]))/32768.f;
            }
        }
    }

    vorbis_analysis_wrote(&s->vd, samples);

    s->samples_in_current_page += samples;
}


/* Returns:
 *   0     No output at this time
 *   >0    Page produced
 *
 * Caller should loop over this to ensure that we don't end up with
 * excessive buffering in libvorbis.
 */
int vorbis_encode_dataout(vorbis_encoder_t *s, ogg_page *og)
{
    ogg_packet op;
    int result;

    if(s->in_header)
    {
        result = ogg_stream_flush(&s->os, og);
        if(result==0) 
        {
            s->in_header = 0;
            return vorbis_encode_dataout(s,og);
        }
        else
            return 1;
    }
    else
    {
        while(vorbis_analysis_blockout(&s->vd, &s->vb)==1)
        {
            vorbis_analysis(&s->vb, NULL);
            vorbis_bitrate_addblock(&s->vb);

            while(vorbis_bitrate_flushpacket(&s->vd, &op)) 
                ogg_stream_packetin(&s->os, &op);
        }

        /* FIXME: Make this threshold configurable.
         * We don't want to buffer too many samples in one page when doing
         * live encoding - that's fine for non-live encoding, but breaks
         * badly when doing things live. 
         * So, we flush the stream if we have too many samples buffered
         */
        if(s->samples_in_current_page > s->samplerate * 2)
        {
            /*LOG_DEBUG1("Forcing flush: Too many samples in current page (%d)",
                    s->samples_in_current_page); */
            result = ogg_stream_flush(&s->os, og);
        }
        else
            result = ogg_stream_pageout(&s->os, og);

        if (result==0) 
	  return 0;

        else /* Page found! */
        {
            s->samples_in_current_page -= ogg_page_granulepos(og) - 
                    s->prevgranulepos;
            s->prevgranulepos = ogg_page_granulepos(og);
            return 1;
        }
    }
}

void vorbis_encode_finish(vorbis_encoder_t *s)
{
    ogg_packet op;
    vorbis_analysis_wrote(&s->vd, 0);

    while(vorbis_analysis_blockout(&s->vd, &s->vb)==1)
    {
        vorbis_analysis(&s->vb, NULL);
        vorbis_bitrate_addblock(&s->vb);
        while(vorbis_bitrate_flushpacket(&s->vd, &op))
            ogg_stream_packetin(&s->os, &op);
    }

}

int vorbis_encode_flush(vorbis_encoder_t *s, ogg_page *og)
{
    int result = ogg_stream_pageout(&s->os, og);

    if(result<=0)
        return 0;
    else
        return 1;
}

void vorbis_encode_clear(vorbis_encoder_t *s)
{
    if(s)
    {
        ogg_stream_clear(&s->os);
        vorbis_block_clear(&s->vb);
        vorbis_dsp_clear(&s->vd);
        vorbis_info_clear(&s->vi);
        free(s);
    }
}


vorbis_encoder_t *vorbis_encode_initialise(int channels, int rate, int managed,
					   int min_br, int nom_br, int max_br, 
					   float quality)
{
  char artist[256];
  char title[256];
  char filename[1024];

  /* int ret; */
  vorbis_comment vc;
  vorbis_encoder_t *s = calloc(1, sizeof(vorbis_encoder_t));
  
  vorbis_info_init(&s->vi);
  if (vorbis_encode_setup_vbr(&s->vi, channels, rate, quality)) {
    stream404_log_debug("Mode initialisation failed: invalid parameters for quality");
  }
  vorbis_encode_ctl(&s->vi, OV_ECTL_RATEMANAGE_AVG, NULL);
  vorbis_encode_setup_init(&s->vi);
  /* if (ret = vorbis_encode_init_vbr(&s->vi, channels, rate, quality); */

  vorbis_comment_init(&vc);
  vorbis_comment_add_tag(&vc,"ENCODER","stream404");
  
  /* get artist and title from metadata */
  artist[0] = '\0';
  title[0] = '\0';
  filename[0] = '\0';
  stream404_metadata_get(artist, 256, title, 256);
  if ((artist[0] == '\0') && (title[0] == '\0')) {
      stream404_metadata_get_file(filename, 1024);
      if (filename != '\0')
	  vorbis_comment_add_tag(&vc,"TITLE",filename);
  } else {
      if (artist[0] != '\0')
	  vorbis_comment_add_tag(&vc,"ARTIST",artist); 
      if (title[0] != '\0')
	  vorbis_comment_add_tag(&vc,"TITLE",title);
  }

  /* set up the analysis state and auxiliary encoding storage */
  vorbis_analysis_init(&s->vd,&s->vi);
  vorbis_block_init(&s->vd,&s->vb);
  
  srand(time(NULL));
  ogg_stream_init(&s->os,rand());
  
  {
    ogg_packet header;
    ogg_packet header_comm;
    ogg_packet header_code;
    
    vorbis_analysis_headerout(&s->vd,&vc,&header,&header_comm,&header_code);
    ogg_stream_packetin(&s->os,&header); /* automatically placed in its own
					    page */
    ogg_stream_packetin(&s->os,&header_comm);
    ogg_stream_packetin(&s->os,&header_code);
    
  }
  s->in_header = 1;
  s->samplerate = rate;
  s->samples_in_current_page = 0;
  s->prevgranulepos = 0;
   
  return s;
}


/* Initialize the reencoding engine in stream404, initialize
 * the liblame structures and be happy */
/* In this version only the vorbis is initialized as lame is for every new 
   song */
void
stream404_reencode_initialize (void)
{
  stream404_stream_t* stream;

  /* are any streams reencoding? */
  for (stream = stream404_config.streams; stream; stream = stream->next)
    if (stream->reencode) {
      stream404_config.reencode = 1;
      /* if and only if for vorbis we initialize here */
      if (stream->type == STREAM404_OUTPUT_VORBIS) {
	//vorbis_reencode_reset(stream);
	/* initialize an encoder for each new stream
	  if (stream->encoder_state)
	  stream404_util_free(stream->encoder_state);
	  if ((stream->encoder_state = 
	  vorbis_encode_initialise(stream->out_numchannels, stream->out_samplerate, 0, -1, -1, -1, 1)) == NULL) {
	  stream404_setup_shutdown();
	}
	*/
      }
      break;
    }

  if (! stream404_config.reencode)
    return;

  stream404_log_debug ("Using LAME version %s", get_lame_version ());
}


/* called by stream404_reencode_reset if server output is an ogg/vorbis stream */
void
vorbis_reencode_reset (input_stream_t *source, stream404_stream_t *stream) 
{
  /* first initialize the resampling process */
  //resample_state_t *resamp = (resample_state_t *)stream->resample_state;
  if ((stream->resample_state = resample_initialise (stream->out_numchannels, source->samplerate, stream->out_samplerate)) == NULL) {
    stream404_setup_shutdown();
  }
  /*
    if (stream->encoder_state)
    stream404_util_free(stream->encoder_state);
  */
  if ((stream->encoder_state =
       vorbis_encode_initialise(stream->out_numchannels, stream->out_samplerate, 1, -1, stream->bitrate, -1, stream->quality*0.1)) == NULL) {
    stream404_setup_shutdown();
  }
}

/* For each song, reset the liblame engine, otherwise it craps out if
 * the bitrate or sample rate changes */
void
stream404_reencode_reset (input_stream_t* source) 
{
  stream404_stream_t* stream;
  lame_global_flags* lame;
  static int init_decoder = 1;

#ifdef HAVE_LAME_DECODE_EXIT
  if (!init_decoder) {
    if (lame_decode_exit () < 0) {
      stream404_log ("LAME: error shutting down decoder");
      stream404_setup_shutdown ();
    }
    init_decoder = 1;
  }
#endif

  if (init_decoder) {
    if (lame_decode_init () < 0) {
      stream404_log ("LAME: error initialising decoder");
      stream404_setup_shutdown ();
    }
    init_decoder = 0;
  }

  for (stream = stream404_config.streams; stream; stream = stream->next) {
    if (! stream->reencode)
      continue;

    switch (stream->type) {
    case STREAM404_OUTPUT_VORBIS:
      /* Do nothing for the vorbis encoder ! */
      /* Or may be we should call encore_initialize */
      vorbis_reencode_reset(source, stream);
      break;
    case STREAM404_OUTPUT_MP3:
    default:
      /* only reset encoder if audio format changes */
      lame = (lame_global_flags*)stream->encoder_state;
      if (lame) {
	if (lame_get_in_samplerate (lame) == source->samplerate)
	  continue;

	lame_close (lame);
      }


      if (! (stream->encoder_state = lame_init ())) {
	stream404_log ("LAME: error resetting encoder.");
	stream404_setup_shutdown ();
      }
      
      lame = (lame_global_flags*)stream->encoder_state;
      
      lame_set_in_samplerate (lame, source->samplerate);
      /* Lame won't reencode mono to stereo for some reason, so we have to
       * duplicate left into right by hand. */
      if (source->channels == 1 && stream->out_numchannels == 1)
	lame_set_num_channels (lame, source->channels);
      
      lame_set_brate (lame, stream->bitrate);
      if (stream->out_numchannels == 1)
	lame_set_mode (lame, MONO);
      if (stream->out_samplerate > 0)
	lame_set_out_samplerate (lame, stream->out_samplerate);
      lame_set_original (lame, 0);
      
      /* added by bedlam */
      lame_set_quality (lame, 2);
      
      /* lame_init_params isn't more specific about the problem */
      if (lame_init_params (lame) < 0) {
	stream404_log ("LAME: error resetting sample rate.");
	lame_close (lame);
	stream404_setup_shutdown ();
      }
      
      /* adjust default stream sample rate to LAME default for lazy reencoding tests */
      if (stream->out_samplerate <= 0)
	stream->out_samplerate = lame_get_out_samplerate (lame);
    }
  }
}

/* If initialized, shutdown the reencoding engine */
void
stream404_reencode_shutdown (void)
{
  stream404_stream_t* stream;

  for (stream = stream404_config.streams; stream; stream = stream->next)
    if (stream->encoder_state) {
      switch (stream->type) {
      case STREAM404_OUTPUT_VORBIS:
	stream404_util_free(stream->encoder_state);
	break;
      case STREAM404_OUTPUT_MP3:
      default:
	lame_close ((lame_global_flags*) stream->encoder_state);
	stream->encoder_state = NULL;
      }
    }
}

/* decode buffer, of length buflen, into left and right. Stream-independent
 * (do this once per chunk, not per stream). Result is number of samples
 * for stream404_reencode_reencode_chunk. */
int
stream404_reencode_decode (unsigned char* buf, size_t blen, size_t olen,
		      int16_t* left, int16_t* right)
{
  return lame_decode (buf, blen, left, right);
}

/* reencode buff, of len buflen, put max outlen reencoded bytes in outbuf */
int
stream404_reencode (stream404_stream_t* stream, int nsamples, int16_t* left,
	       int16_t* right, unsigned char *outbuf, int outlen)
{
  lame_global_flags* lame = (lame_global_flags*) stream->encoder_state;

  return lame_encode_buffer (lame, left, right, nsamples, outbuf, outlen);
}

/* At EOF of each file, flush the liblame buffers and get some extra candy */
int
stream404_reencode_flush (stream404_stream_t* stream, unsigned char *outbuf,
		     int maxlen)
{
  lame_global_flags* lame = (lame_global_flags*) stream->encoder_state;
  int rc;

  rc = lame_encode_flush_nogap (lame, (char *)outbuf, maxlen);

  return rc;
}









/*--------------------------------- resampling part -------------------------*/

/* Some systems don't define this */
#ifndef M_PI
#define M_PI       3.14159265358979323846 
#endif

static int hcf(int arg1, int arg2)
{
    int mult = 1;

    while (~(arg1 | arg2) & 1)
        arg1 >>= 1, arg2 >>= 1, mult <<= 1;

    while (arg1 > 0)
    {
        if (~(arg1 & arg2) & 1)
        {
            arg1 >>= (~arg1 & 1);
            arg2 >>= (~arg2 & 1);
        }
        else if (arg1 < arg2)
            arg2 = (arg2 - arg1) >> 1;
        else
            arg1 = (arg1 - arg2) >> 1;
    }

    return arg2 * mult;
}


static void filt_sinc(float *dest, int N, int step, double fc, double gain, int width)
{
    double s = fc / step;
    int mid, x;
    float *endpoint = dest + N,
        *base = dest,
        *origdest = dest;
    
    assert(width <= N);

    if ((N & 1) == 0)
    {
        *dest = 0.0;
        dest += width;
        if (dest >= endpoint)
            dest = ++base;
        N--;
    }

    mid = N / 2;
    x = -mid;

    while (N--)
    {
        *dest = (x ? sin(x * M_PI * s) / (x * M_PI) * step : fc) * gain;
        x++;
        dest += width;
        if (dest >= endpoint)
            dest = ++base;
    }
    assert(dest == origdest + width);
}


static double I_zero(double x)
{
    int n = 0;
    double u = 1.0,
        s = 1.0,
        t;

    do
    {
        n += 2;
        t = x / n;
        u *= t * t;
        s += u;
    } while (u > 1e-21 * s);

    return s;
}


static void win_kaiser(float *dest, int N, double alpha, int width)
{
    double I_alpha, midsq;
    int x;
    float *endpoint = dest + N,
        *base = dest,
        *origdest = dest;

    assert(width <= N);

    if ((N & 1) == 0)
    {
        *dest = 0.0;
        dest += width;
        if (dest >= endpoint)
            dest = ++base;
        N--;
    }

    x = -(N / 2);
    midsq = (double)(x - 1) * (double)(x - 1);
    I_alpha = I_zero(alpha);

    while (N--)
    {
        *dest *= I_zero(alpha * sqrt(1.0 - ((double)x * (double)x) / midsq)) / I_alpha;
        x++;
        dest += width;
        if (dest >= endpoint)
            dest = ++base;
    }
    assert(dest == origdest + width);
}


int resampler_init(resampler_state_t *state, int channels, int outfreq, int infreq, resampler_parameter op1, ...)
{
    double beta = 16.0,
        cutoff = 0.80,
        gain = 1.0;
    int taps = 45;

    int factor;

    assert(state);
    assert(channels > 0);
    assert(outfreq > 0);
    assert(infreq > 0);
    assert(taps > 0);

    if (state == NULL || channels <= 0 || outfreq <= 0 || infreq <= 0 || taps <= 0)
        return -1;

    if (op1 != RES_END)
    {
        va_list argp;
        va_start(argp, op1);
        do
        {
            switch (op1)
            {
            case RES_GAIN:
                gain = va_arg(argp, double);
                break;

            case RES_CUTOFF:
                cutoff = va_arg(argp, double);
                assert(cutoff > 0.01 && cutoff <= 1.0);
                break;

            case RES_TAPS:
                taps = va_arg(argp, int);
                assert(taps > 2 && taps < 1000);
                break;
                
            case RES_BETA:
                beta = va_arg(argp, double);
                assert(beta > 2.0);
                break;
            default:
                assert("arglist" == "valid");
                return -1;
            }
            op1 = va_arg(argp, resampler_parameter);
        } while (op1 != RES_END);
        va_end(argp);
    }

    factor = hcf(infreq, outfreq);
    outfreq /= factor;
    infreq /= factor;

    /* adjust to rational values for downsampling */
    if (outfreq < infreq)
    {
        /* push the cutoff frequency down to the output frequency */
        cutoff = cutoff * outfreq / infreq; 

        /* compensate for the sharper roll-off requirement
         * by using a bigger hammer */
        taps = taps * infreq/outfreq;
    }

    assert(taps >= (infreq + outfreq - 1) / outfreq);

    if ((state->table = calloc(outfreq * taps, sizeof(float))) == NULL)
        return -1;
    if ((state->pool = calloc(channels * taps, sizeof(SAMPLE))) == NULL)
    {
        free(state->table);
        state->table = NULL;
        return -1;
    }

    state->poolfill = taps / 2 + 1;
    state->channels = channels;
    state->outfreq = outfreq;
    state->infreq = infreq;
    state->taps = taps;
    state->offset = 0;

    filt_sinc(state->table, outfreq * taps, outfreq, cutoff, gain, taps);
    win_kaiser(state->table, outfreq * taps, beta, taps);

    return 0;
}


static SAMPLE sum(float const *scale, int count, SAMPLE const *source, SAMPLE const *trigger, SAMPLE const *reset, int srcstep)
{
    float total = 0.0;

    while (count--)
    {
        total += *source * *scale;

        if (source == trigger)
            source = reset, srcstep = 1;
        source -= srcstep;
        scale++;
    }

    return total;
}


static int push(resampler_state_t const * const state, SAMPLE *pool, int * const poolfill, int * const offset, SAMPLE *dest, int dststep, SAMPLE const *source, int srcstep, size_t srclen)
{
    SAMPLE    * const destbase = dest,
        *poolhead = pool + *poolfill,
        *poolend = pool + state->taps,
        *newpool = pool;
    SAMPLE const *refill, *base, *endpoint;
    int    lencheck;

    
    assert(state);
    assert(pool);
    assert(poolfill);
    assert(dest);
    assert(source);

    assert(state->poolfill != -1);
    
    lencheck = resampler_push_check(state, srclen);

    /* fill the pool before diving in */
    while (poolhead < poolend && srclen > 0)
    {
        *poolhead++ = *source;
        source += srcstep;
        srclen--;
    }

    if (srclen <= 0)
        return 0;

    base = source;
    endpoint = source + srclen * srcstep;

    while (source < endpoint)
    {
        *dest = sum(state->table + *offset * state->taps, state->taps, source, base, poolend, srcstep);
        dest += dststep;
        *offset += state->infreq;
        while (*offset >= state->outfreq)
        {
            *offset -= state->outfreq;
            source += srcstep;
        }
    }

    assert(dest == destbase + lencheck * dststep);

    /* pretend that source has that underrun data we're not going to get */
    srclen += (source - endpoint) / srcstep;

    /* if we didn't get enough to completely replace the pool, then shift things about a bit */
    if (srclen < state->taps)
    {
        refill = pool + srclen;
        while (refill < poolend)
            *newpool++ = *refill++;

        refill = source - srclen * srcstep;
    }
    else
        refill = source - state->taps * srcstep;

    /* pull in fresh pool data */
    while (refill < endpoint)
    {
        *newpool++ = *refill;
        refill += srcstep;
    }

    assert(newpool > pool);
    assert(newpool <= poolend);

    *poolfill = newpool - pool;

    return (dest - destbase) / dststep;
}


 int resampler_push_max_input(resampler_state_t const * const state, size_t maxoutput)
{
    return maxoutput * state->infreq / state->outfreq;
}


int resampler_push_check(resampler_state_t const * const state, size_t srclen)
{
    if (state->poolfill < state->taps)
        srclen -= state->taps - state->poolfill;

    return (srclen * state->outfreq - state->offset + state->infreq - 1) / state->infreq;
}


int resampler_push(resampler_state_t *state, SAMPLE **dstlist, SAMPLE const **srclist, size_t srclen)
{
    int result = -1, poolfill = -1, offset = -1, i;

    assert(state);
    assert(dstlist);
    assert(srclist);
    assert(state->poolfill >= 0);

    for (i = 0; i < state->channels; i++)
    {
        poolfill = state->poolfill;
        offset = state->offset;
        result = push(state, state->pool + i * state->taps, &poolfill, &offset, dstlist[i], 1, srclist[i], 1, srclen);
    }
    state->poolfill = poolfill;
    state->offset = offset;

    return result;
}


int resampler_push_interleaved(resampler_state_t *state, SAMPLE *dest, SAMPLE const *source, size_t srclen)
{
    int result = -1, poolfill = -1, offset = -1, i;
    
    assert(state);
    assert(dest);
    assert(source);
    assert(state->poolfill >= 0);

    for (i = 0; i < state->channels; i++)
    {
        poolfill = state->poolfill;
        offset = state->offset;
        result = push(state, state->pool + i * state->taps, &poolfill, &offset, dest + i, state->channels, source + i, state->channels, srclen);
    }
    state->poolfill = poolfill;
    state->offset = offset;

    return result;
}


int resampler_drain(resampler_state_t *state, SAMPLE **dstlist)
{
    SAMPLE *tail;
    int result = -1, poolfill = -1, offset = -1, i;

    assert(state);
    assert(dstlist);
    assert(state->poolfill >= 0);

    if ((tail = calloc(state->taps, sizeof(SAMPLE))) == NULL)
        return -1;

    for (i = 0; i < state->channels; i++)
    {
        poolfill = state->poolfill;
        offset = state->offset;
        result = push(state, state->pool + i * state->taps, &poolfill, &offset, dstlist[i], 1, tail, 1, state->taps / 2 - 1);
    }
        
    free(tail);

    state->poolfill = -1;

    return result;
}


int resampler_drain_interleaved(resampler_state_t *state, SAMPLE *dest)
{
    SAMPLE *tail;
    int result = -1, poolfill = -1, offset = -1, i;

    assert(state);
    assert(dest);
    assert(state->poolfill >= 0);

    if ((tail = calloc(state->taps, sizeof(SAMPLE))) == NULL)
        return -1;

    for (i = 0; i < state->channels; i++)
    {
        poolfill = state->poolfill;
        offset = state->offset;
        result = push(state, state->pool + i * state->taps, &poolfill, &offset, dest + i, state->channels, tail, 1, state->taps / 2 - 1);
    }
        
    free(tail);

    state->poolfill = -1;

    return result;
}


void resampler_clear(resampler_state_t *state)
{
    assert(state);
    assert(state->table);
    assert(state->pool);

    free(state->table);
    free(state->pool);
    memset(state, 0, sizeof(*state));
}


/*----------------------------- audio.c ------------------------------------*/


downmix_state_t *downmix_initialise(void) {
    downmix_state_t *state = calloc(1, sizeof(downmix_state_t));

    stream404_log_debug("Enabling stereo->mono downmixing");

    return state;
}

void downmix_clear(downmix_state_t *s) {
    if(s) {
        if (s->buffer)
            free(s->buffer);
        free(s);
    }
}

void downmix_buffer_float(downmix_state_t *s, float **buf, int samples)
{
    int i;

    if(samples > s->buflen) {
        void *tmp = realloc(s->buffer, samples * sizeof(float));
        if (tmp==NULL)
            return;
        s->buffer = tmp;
        s->buflen = samples;
    }

    for(i=0; i < samples; i++) {
        s->buffer[i] = (buf[0][i] + buf[1][i])*0.5;
    }
    
}


void downmix_buffer(downmix_state_t *s, signed char *buf, int len, int be)
{
    int samples = len/4;
    int i;

    if(samples > s->buflen) {
        void *tmp = realloc(s->buffer, samples * sizeof(float));
        if (tmp==NULL)
            return;
        s->buffer = tmp;
        s->buflen = samples;
    }

    if(be) {
        for(i=0; i < samples; i++) {
            s->buffer[i] = (((buf[4*i]<<8) | (buf[4*i + 1]&0xff)) +
                           ((buf[4*i + 2]<<8) | (buf[4*i + 3]&0xff)))/65536.f;
        }
    }
    else {
        for(i=0; i < samples; i++) {
            s->buffer[i] = (((buf[4*i + 1]<<8) | (buf[4*i]&0xff)) +
                           ((buf[4*i + 3]<<8) | (buf[4*i + 2]&0xff)))/65536.f;
        }
    }
}

resample_state_t *resample_initialise(int channels, int infreq, int outfreq)
{
    resample_state_t *state = calloc(1, sizeof(resample_state_t));
    int failed = 1;

    do 
    {
        if (state==NULL)
            break;
        if (resampler_init(&state->resampler, channels, outfreq, infreq, RES_END)) {
            stream404_log_debug("Couldn't initialise resampler to specified frequency");
            return NULL;
        }

        if ((state->buffers = calloc(channels, sizeof(float *))) == NULL)
            break;
        if ((state->convbuf = calloc(channels, sizeof(float *))) == NULL)
            break;
        failed = 0;
    }
    while (0); /* not a loop */
        
    if (failed)
    {
        stream404_log_debug("Couldn't initialise resampler due to memory allocation failure");
        resample_clear (state);
        return NULL;
    }
    state->channels = channels;

    return state;
}

void resample_clear(resample_state_t *s)
{
    int c;

    if(s) {
        if(s->buffers) {
            for(c=0; c<s->channels; c++)
                if (s->buffers[c])
                    free(s->buffers[c]);
            free(s->buffers);
        }
        if(s->convbuf) {
            for(c=0; c<s->channels; c++)
                if (s->convbuf[c])
                    free(s->convbuf[c]);
            free(s->convbuf);
        }
        resampler_clear(&s->resampler);
        free(s);
    }
}

void resample_buffer(resample_state_t *s, signed char *buf, int buflen, int be)
{
    int c,i;
    buflen /= 2*s->channels; /* bytes -> samples conversion */

    if(s->convbuflen < buflen) {
        s->convbuflen = buflen;
        for(c=0; c < s->channels; c++)
            s->convbuf[c] = realloc(s->convbuf[c], buflen * sizeof(float));
    }

    if(be) {
        for(i=0; i < buflen; i++) {
            for(c=0; c < s->channels; c++) {
                s->convbuf[c][i] = ((buf[2*(i*s->channels + c)]<<8) |
                                    (0x00ff&(int)buf[2*(i*s->channels + c)+1]))/
                    32768.f;
            }
        }
    }
    else {
        for(i=0; i < buflen; i++) {
            for(c=0; c < s->channels; c++) {
                s->convbuf[c][i] = ((buf[2*(i*s->channels + c) + 1]<<8) |
                                    (0x00ff&(int)buf[2*(i*s->channels + c)]))/
                    32768.f;
            }
        }
    }

    resample_buffer_float(s, s->convbuf, buflen);
}

void resample_buffer_float(resample_state_t *s, float **buf, int buflen)
{
    int c;
    int res;

    s->buffill = resampler_push_check(&s->resampler, buflen);
    if(s->buffill <= 0) {
        stream404_log_debug("Fatal reencoding error: resampler_push_check returned %d", 
                s->buffill);
    }

    if(s->bufsize < s->buffill) {
        s->bufsize = s->buffill;
        for(c=0; c<s->channels; c++)
            s->buffers[c] = realloc(s->buffers[c], s->bufsize * sizeof(float));
    }

    if((res = resampler_push(&s->resampler, s->buffers, (float const **)buf, buflen))
            != s->buffill) {
        stream404_log_debug("Internal error in resampling: returned number of samples %d"
                   ", expected %d", res, s->buffill);
        s->buffill = res;
        return;
    }

}

void resample_finish(resample_state_t *s)
{
    int ret;

    if(!s->buffers[0])
        return;

    ret = resampler_drain(&s->resampler, s->buffers);

    if(ret > s->bufsize) {
        stream404_log_debug("Fatal error in resampler: buffers too small");
        return;
    }

    s->buffill = ret;
}
