/* $Id: setup.c,v 1.3 2005/07/18 18:16:44 bedlam Exp $ */
/* setup.c
 * - Functions for initialization in ices
 * Copyright (c) 2000 Alexander Havng
 * Copyright (c) 2002-4 Brendan Cully <brendan@xiph.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "definitions.h"
#include "metadata.h"

/* Local function declarations */
static void stream404_setup_parse_options (stream404_config_t *stream404_config);
static void stream404_setup_parse_defaults (stream404_config_t *stream404_config);
static void stream404_setup_parse_sql_defaults (sql_t *sql);
#ifdef HAVE_LIBXML
static void stream404_setup_parse_config_file (stream404_config_t *stream404_config, const char *configfile);
#endif
static void stream404_setup_parse_command_line (stream404_config_t *stream404_config, char **argv, int argc);
static void stream404_setup_parse_command_line_for_new_configfile (stream404_config_t *stream404_config, char **argv, int argc);
static void stream404_setup_activate_libshout_changes (const stream404_config_t *stream404_config);
static void stream404_setup_usage (void);
static void stream404_setup_version (void);
static void stream404_setup_update_pidfile (int stream404pid);
static void stream404_setup_daemonize (void);
static void stream404_free_all (stream404_config_t *stream404_config);

extern stream404_config_t stream404_config;

/* Global function definitions */

/* Top level initialization function for stream404.
 * It will parse options, initialize modules,
 * and if requested, become a daemon. */
void
stream404_setup_initialize (void)
{
  stream404_stream_t* stream;
  stream404_plugin_t* plugin;
  
  shout_init();

  /* Setup signal handlers */
  stream404_signals_setup ();

  /* Parse the options in the config file, and the command line */
  stream404_setup_parse_options (&stream404_config);

  if (stream404_config.daemon)
    stream404_setup_daemonize ();

  /* Open logfiles */
  stream404_log_initialize ();

  /* Initialize the libshout structure */
  for (stream = stream404_config.streams; stream; stream = stream->next) {
    if (!(stream->conn = shout_new ())) {
      stream404_log ("Could not create shout interface");
      stream404_setup_shutdown ();
    }
  }

  stream404_setup_activate_libshout_changes (&stream404_config);

  /* Initialize the playlist handler */
  stream404_playlist_initialize ();

  /* Initialize liblame for reeencoding */
  stream404_reencode_initialize ();
  
  while (stream404_config.plugins && stream404_config.plugins->init() < 0)
    stream404_config.plugins = stream404_config.plugins->next;

  for (plugin = stream404_config.plugins; plugin && plugin->next; plugin = plugin->next)
    if (plugin->next->init() < 0)
      plugin->next = plugin->next->next;
  
  stream404_log_debug("Startup complete\n");
}

/* Top level stream404 shutdown function.
 * This is the _only_ way out of here */
void
stream404_setup_shutdown (void)
{
  stream404_stream_t* stream;
  stream404_plugin_t* plugin;

  /* Tell libshout to disconnect from server */
  for (stream = stream404_config.streams; stream; stream = stream->next)
    if (stream->conn)
      shout_close (stream->conn);

  for (plugin = stream404_config.plugins; plugin; plugin = plugin->next)
    plugin->shutdown();

  /* Order the reencoding engine to shutdown */
  stream404_reencode_shutdown ();

  /* Tell the playlist module to shutdown and cleanup */
  stream404_playlist_shutdown ();

  /* Cleanup the cue file (the cue module has no init yet) */
  stream404_cue_shutdown ();

  /* Make sure we're not leaving any memory allocated around when
   * we exit. This makes it easier to find memory leaks, and 
   * some systems actually don't clean up that well */
  stream404_free_all (&stream404_config);
	
  /* Let the log and console know we went down ok */
  stream404_log ("Stream404 Exiting...");

  /* Close logfiles */
  stream404_log_shutdown ();
	
  shout_shutdown();

  /* Down and down we go... */
  exit (1);
}

/* Local function definitions */

/* Top level option parsing function.
 * Sets of options object (stream404_config), with:
 * - Hardcoded defaults
 * - Configfile settings
 * - Command line options */
static void
stream404_setup_parse_options (stream404_config_t *stream404_config)
{
  /* Get default values for the settings */
  stream404_setup_parse_defaults (stream404_config);

  /* Look for given configfile on the commandline */
  stream404_setup_parse_command_line_for_new_configfile (stream404_config, stream404_util_get_argv(), stream404_util_get_argc ());

#ifdef HAVE_LIBXML
  /* Parse the configfile */
  stream404_setup_parse_config_file (stream404_config, stream404_config->configfile);
#endif
	
  /* Parse the commandline */
  stream404_setup_parse_command_line (stream404_config, stream404_util_get_argv(), stream404_util_get_argc());
}

/* Function for placing hardcoded defaults in the 
 * options object (stream404_config) */
static void
stream404_setup_parse_defaults (stream404_config_t *stream404_config)
{
  stream404_config->configfile = stream404_util_strdup (STREAM404_DEFAULT_CONFIGFILE);
  stream404_config->daemon = STREAM404_DEFAULT_DAEMON;
  stream404_config->base_directory = stream404_util_strdup (STREAM404_DEFAULT_BASE_DIRECTORY);
  stream404_config->verbose = STREAM404_DEFAULT_VERBOSE;
  stream404_config->reencode = 0;

  stream404_config->pm.playlist_file =
    stream404_util_strdup (STREAM404_DEFAULT_PLAYLIST_FILE);
  stream404_config->pm.module = 
      stream404_util_strdup (STREAM404_DEFAULT_MODULE);
  stream404_config->pm.randomize = STREAM404_DEFAULT_RANDOMIZE_PLAYLIST;
  stream404_config->pm.playlist_type = STREAM404_DEFAULT_PLAYLIST_TYPE;
  
  /* by default there is no backup playlist */
  stream404_config->pm_backup.playlist_type = stream404_playlist_null_e;
  stream404_config->pm_backup.playlist_file =
      stream404_util_strdup (STREAM404_DEFAULT_PLAYLIST_FILE);
  stream404_config->pm_backup.module = 
      stream404_util_strdup (STREAM404_DEFAULT_MODULE);
  stream404_config->pm_backup.randomize = STREAM404_DEFAULT_RANDOMIZE_PLAYLIST;

  stream404_config->streams = (stream404_stream_t*) malloc (sizeof (stream404_stream_t));

  stream404_setup_parse_sql_defaults (&stream404_config->sql);

  stream404_setup_parse_stream_defaults (stream404_config->streams);
}

static void 
stream404_setup_parse_sql_defaults (sql_t *sql)
{
    sql->hostname = NULL;
    sql->usocket = strdup(STREAM404_DEFAULT_SQL_USOCKET);
    sql->port = STREAM404_DEFAULT_SQL_PORT;
    sql->sqltype = strdup(STREAM404_DEFAULT_SQL_SQLTYPE);
    sql->user = strdup(STREAM404_DEFAULT_SQL_USER);
    sql->password = strdup(STREAM404_DEFAULT_SQL_PASSWORD);
    sql->database = strdup(STREAM404_DEFAULT_SQL_DATABASE);
    sql->table = strdup(STREAM404_DEFAULT_SQL_TABLE);
    sql->urg_table = strdup(STREAM404_DEFAULT_SQL_URGTABLE);
    sql->currentstream_table = strdup(STREAM404_DEFAULT_SQL_CURRENTSTREAMTABLE);
    sql->streamlog_table = strdup(STREAM404_DEFAULT_SQL_STREAMLOGTABLE);
    sql->metadata_table = strdup(STREAM404_DEFAULT_SQL_METADATATABLE);
}

/* Place hardcoded defaults into an stream404_stream_t object */
void
stream404_setup_parse_stream_defaults (stream404_stream_t* stream)
{
  stream->conn = NULL;
  stream->host = stream404_util_strdup (STREAM404_DEFAULT_HOST);
  stream->port = STREAM404_DEFAULT_PORT;
  stream->password = stream404_util_strdup (STREAM404_DEFAULT_PASSWORD);
  stream->protocol = STREAM404_DEFAULT_PROTOCOL;

  stream->mount = stream404_util_strdup (STREAM404_DEFAULT_MOUNT);
  stream->dumpfile = NULL;
  
  stream->name = stream404_util_strdup (STREAM404_DEFAULT_NAME);
  stream->genre = stream404_util_strdup (STREAM404_DEFAULT_GENRE);
  stream->description = stream404_util_strdup (STREAM404_DEFAULT_DESCRIPTION);
  stream->url = stream404_util_strdup (STREAM404_DEFAULT_URL);
  stream->ispublic = STREAM404_DEFAULT_ISPUBLIC;

  stream->bitrate = STREAM404_DEFAULT_BITRATE;
  stream->quality = STREAM404_DEFAULT_QUALITY;
  stream->reencode = STREAM404_DEFAULT_REENCODE;
  stream->out_numchannels = -1;
  stream->out_samplerate = -1;

  stream->encoder_state = NULL;
  stream->connect_delay = 0;
  
  stream->type = STREAM404_DEFAULT_OUTPUT_TYPE;
  
  stream->next = NULL;
}

/* Frees stream404_stream_t data (but not the object itself) */
static void
stream404_setup_free_stream (stream404_stream_t* stream)
{
  if (stream->conn)
    shout_free (stream->conn);
  stream404_util_free (stream->host);
  stream404_util_free (stream->password);

  stream404_util_free (stream->mount);
  stream404_util_free (stream->dumpfile);

  stream404_util_free (stream->name);
  stream404_util_free (stream->genre);
  stream404_util_free (stream->description);
  stream404_util_free (stream->url);
}

/* Function to free() all allocated memory when stream404 shuts down. */
static void
stream404_free_all (stream404_config_t *stream404_config)
{
  stream404_stream_t *stream, *next;

  stream404_util_free (stream404_config->configfile);
  stream404_util_free (stream404_config->base_directory);

  stream404_util_free (stream404_config->pm.playlist_file);
  stream404_util_free (stream404_config->pm.module);

  for (stream = stream404_config->streams; stream; stream = next)
  {
    next = stream->next;

    stream404_setup_free_stream (stream);
    stream404_util_free (stream);
  }
}

#ifdef HAVE_LIBXML
/* Tell the xml module to parse the config file. */
static void
stream404_setup_parse_config_file (stream404_config_t *stream404_config, const char *configfile)
{
  char namespace[1024];
  const char *realname = NULL;
  int ret;

  if (stream404_util_verify_file (configfile)) {
    realname = configfile;
  } else {
    sprintf (namespace, "%s/%s", STREAM404_ETCDIR, configfile);
    if (stream404_util_verify_file (namespace))
      realname = &namespace[0];
  }
	
  if (realname) {
    ret = stream404_xml_parse_config_file (stream404_config, realname);

    if (ret == -1) {
      /* ret == -1 means we have no libxml support */
      stream404_log_debug ("%s", stream404_log_get_error ());
    } else if (ret == 0) {
      /* A real error */
      stream404_log ("%s", stream404_log_get_error ());
    }
  }
}
#endif

/* This function looks through the command line options for a new
 * configfile. */
static void
stream404_setup_parse_command_line_for_new_configfile (stream404_config_t *stream404_config, char **argv, int argc)
{
  int arg;
  char *s;

  arg = 1;

  while (arg < argc) {
    s = argv[arg];

    if (s[0] == '-') {
      switch (s[1]) {
      case 'c':
	arg++;
	if (stream404_config->configfile)
	  stream404_util_free (stream404_config->configfile);
	stream404_config->configfile = stream404_util_strdup (argv[arg]);
#ifndef HAVE_LIBXML
	fprintf (stderr, "Cannot use config file (no XML support).\n");
	stream404_setup_shutdown ();
#endif
	break;
      }
    }
    arg++;
  }
}

/* This function parses the command line options */
static void
stream404_setup_parse_command_line (stream404_config_t *stream404_config, char **argv,
			       int argc)
{
  int arg;
  char *s;
  stream404_stream_t* stream = stream404_config->streams;
  /* each -m option creates a new stream, subsequent options are applied
   * to it. */
  int nstreams = 1;

  arg = 1;

  while (arg < argc) {
    s = argv[arg];

    if (s[0] == '-') {
      if ((strchr ("BRrsVv", s[1]) == NULL) && arg >= (argc - 1)) {
	fprintf (stderr, "Option %c requires an argument!\n", s[1]);
	stream404_setup_usage ();
	stream404_setup_shutdown ();
	return;
      }

      switch (s[1]) {
        case 'B':
	  stream404_config->daemon = 1;
	  break;
        case 'b':
	  arg++;
	  stream->bitrate = atoi (argv[arg]);
	  break;
        case 'C':
	  arg++;
	  if (atoi(argv[arg]) > 0)
	      /* TODO: stack plugins */
	      /* XXX we cant get CrossfadeFileMin through command line
		 this value is set to 0
	       */
	      stream404_config->plugins = crossfade_plugin(atoi(argv[arg]), 0);
        case 'c':
	  arg++;
	  break;
        case 'd':
	  arg++;
	  stream404_util_free (stream->description);
	  stream->description = stream404_util_strdup (argv[arg]);
	  break;
        case 'D':
	  arg++;
	  stream404_util_free (stream404_config->base_directory);
	  stream404_config->base_directory = stream404_util_strdup (argv[arg]);
	  break;
        case 'F':
	  arg++;
	  stream404_util_free (stream404_config->pm.playlist_file);
	  stream404_config->pm.playlist_file = stream404_util_strdup (argv[arg]);
	  break;
        case 'f':
	  arg++;
	  stream404_util_free (stream->dumpfile);
	  stream->dumpfile = stream404_util_strdup (argv[arg]);
	  break;
        case 'g':
	  arg++;
	  stream404_util_free (stream->genre);
	  stream->genre = stream404_util_strdup (argv[arg]);
	  break;
        case 'h':
	  arg++;
	  stream404_util_free (stream->host);
	  stream->host = stream404_util_strdup (argv[arg]);
	  break;
        case 'H':
	  arg++;
	  stream->out_samplerate = atoi (argv[arg]);
	  break;
        case 'M':
	  arg++;
	  stream404_util_free (stream404_config->pm.module);
	  stream404_config->pm.module = stream404_util_strdup (argv[arg]);
	  break;
        case 'm':
	  arg++;
	  if (nstreams > 1) {
	    stream->next =
	      (stream404_stream_t*) malloc (sizeof (stream404_stream_t));
	    stream = stream->next;
	    stream404_setup_parse_stream_defaults (stream);
	  }
	  stream404_util_free (stream->mount);
	  stream->mount = stream404_util_strdup (argv[arg]);
	  nstreams++;
	  break;
        case 'N':
	  arg++;
	  stream->out_numchannels = atoi (argv[arg]);
	  break;
        case 'n':
	  arg++;
	  stream404_util_free (stream->name);
	  stream->name = stream404_util_strdup (argv[arg]);
	  break;
        case 'P':
	  arg++;
	  stream404_util_free (stream->password);
	  stream->password = stream404_util_strdup (argv[arg]);
	  break;
        case 'p':
	  arg++;
	  stream->port = atoi (argv[arg]);
	  break;
        case 'R':
	  stream->reencode = 1;
	  break;
        case 'r':
	  stream404_config->pm.randomize = 1;
	  break;
        case 'S':
	  arg++;
	  if (strcmp (argv[arg], "python") == 0)
	    stream404_config->pm.playlist_type = stream404_playlist_python_e;
	  else if (strcmp (argv[arg], "perl") == 0)
	    stream404_config->pm.playlist_type = stream404_playlist_perl_e;
	  else 
	    stream404_config->pm.playlist_type = stream404_playlist_builtin_e;
	  break;
        case 's':
	  stream->ispublic = 0;
	  break;
        case 't':
	  arg++;
	  if (!strcmp (argv[arg], "http"))
	    stream->protocol = http_protocol_e;
	  else if (!strcmp (argv[arg], "xaudiocast"))
	    stream->protocol = xaudiocast_protocol_e;
	  else if (!strcmp (argv[arg], "icy"))
	    stream->protocol = icy_protocol_e;
	  else {
	    fprintf (stderr, "Unknown protocol %s. Use 'http', 'xaudiocast' or 'icy'.\n", argv[arg]);
	    stream404_setup_shutdown ();
	  }
	  break;
        case 'u':
	  arg++;
	  stream404_util_free (stream->url);
	  stream->url = stream404_util_strdup (argv[arg]);
	  break;
        case 'V':
	  stream404_setup_version ();
	  exit (0);
        case 'v':
	  stream404_config->verbose = 1;
	  break;
        default:
	  stream404_setup_usage ();
	  break;
      }
    }
    arg++;
  }
}

/* This function takes all the new configuration and copies it to the
   libshout object. */
static void
stream404_setup_activate_libshout_changes (const stream404_config_t *stream404_config)
{
  stream404_stream_t* stream;
  shout_t* conn;
  int streamno = 0;
  char useragent[64];
  char bitrate[8];
  char quality[8];

  snprintf(useragent, sizeof(useragent), "stream404/" VERSION " libshout/%s",
    shout_version(NULL, NULL, NULL));

  for (stream = stream404_config->streams; stream; stream = stream->next) {
    conn = stream->conn;

    shout_set_host (conn, stream->host);
    shout_set_port (conn, stream->port);
    shout_set_password (conn, stream->password);
    if (stream->type == STREAM404_OUTPUT_VORBIS) 
      shout_set_format (conn, SHOUT_FORMAT_VORBIS);
    else 
      shout_set_format (conn, SHOUT_FORMAT_MP3);
    if (stream->protocol == icy_protocol_e)
      shout_set_protocol(conn, SHOUT_PROTOCOL_ICY);
    else if (stream->protocol == http_protocol_e)
      shout_set_protocol(conn, SHOUT_PROTOCOL_HTTP);
    else
      shout_set_protocol(conn, SHOUT_PROTOCOL_XAUDIOCAST);
    if (stream->dumpfile)
      shout_set_dumpfile(conn, stream->dumpfile);
    shout_set_name (conn, stream->name);
    shout_set_url (conn, stream->url);
    shout_set_genre (conn, stream->genre);
    shout_set_description (conn, stream->description);

    if (stream->type == STREAM404_OUTPUT_VORBIS) {
      snprintf(quality, sizeof(quality), "%2.2f", stream->quality);
      shout_set_audio_info (conn, SHOUT_AI_QUALITY, quality);
    } else {
      snprintf(bitrate, sizeof(bitrate), "%d", stream->bitrate);
      shout_set_audio_info (conn, SHOUT_AI_BITRATE, bitrate);
    }

    shout_set_public (conn, stream->ispublic);
    shout_set_mount (conn, stream->mount);
    shout_set_agent (conn, useragent);

    stream404_log_debug ("Sending following information to libshout:");
    stream404_log_debug ("Stream: %d", streamno);
    stream404_log_debug ("Host: %s:%d (protocol: %s) (format: %s)", shout_get_host (conn),
                    shout_get_port (conn),
		    stream->protocol == icy_protocol_e ? "icy" :
		    stream->protocol == http_protocol_e ? "http" : "xaudiocast",
		    stream->type == STREAM404_OUTPUT_VORBIS ? "vorbis" : "mp3");
    stream404_log_debug ("Mount: %s, Password: %s", shout_get_mount (conn), shout_get_password (conn));
    stream404_log_debug ("Name: %s\tURL: %s", shout_get_name (conn), shout_get_url(conn));
    stream404_log_debug ("Genre: %s\tDesc: %s", shout_get_genre (conn),
		    shout_get_description (conn));
    stream404_log_debug ("Bitrate: %s\tQuality: %s\tPublic: %d", shout_get_audio_info (conn, SHOUT_AI_BITRATE), shout_get_audio_info (conn, SHOUT_AI_QUALITY),
		    shout_get_public (conn));
    stream404_log_debug ("Dump file: %s", stream404_util_nullcheck (shout_get_dumpfile (conn)));
    streamno++;
  }
}

/* Display all command line options for stream404 */
static void
stream404_setup_usage (void)
{
  printf ("This is Stream404 from stream404 " VERSION " ");
  printf ("patched by bedlam\n");
  printf ("stream404 <options>\n");
  printf ("Options:\n");
  printf ("\t-B (Background (daemon mode))\n");
  /*
  printf ("\t-b <stream bitrate>\n");
  printf ("\t-C <crossfade seconds>\n");
  */
  printf ("\t-c <configfile>\n");
  printf ("\t-D <base directory>\n");
  /*
  printf ("\t-d <stream description>\n");
  printf ("\t-f <dumpfile on server>\n");
  */
  printf ("\t-F <playlist>\n");
  /*
  printf ("\t-g <stream genre>\n");
  printf ("\t-h <host>\n");
  printf ("\t-i (use icy headers)\n");
  printf ("\t-M <interpreter module>\n");
  printf ("\t-m <mountpoint>\n");
  printf ("\t-n <stream name>\n");
  printf ("\t-p <port>\n");
  printf ("\t-P <password>\n");
  printf ("\t-R (activate reencoding)\n");
  printf ("\t-r (randomize playlist)\n");
  printf ("\t-s (private stream)\n");
  printf ("\t-S <perl|python|builtin>\n");
  printf ("\t-t <http|xaudiocast|icy>\n");
  printf ("\t-u <stream url>\n");
  */
  printf ("\t-v (verbose output)\n");
  /*
  printf ("\t-H <reencoded sample rate>\n");
  printf ("\t-N <reencoded number of channels>\n");
  */
}

/* display version information */
static void
stream404_setup_version (void)
{
  printf ("Stream404 - stream404 " VERSION "\nFeatures: "
  "LAME "
#ifdef HAVE_LIBPERL
  "Perl "
#endif
#ifdef HAVE_LIBPYTHON
  "python "
#endif
#ifdef HAVE_LIBXML
  "libxml "
#endif
  "Vorbis "
#ifdef HAVE_LIBFLAC
  "FLAC "
#endif
#ifdef HAVE_LIBFAAD
  "MP4"
#endif

  "\n"
  "System configuration file: " STREAM404_ETCDIR "/stream404.conf\n"
#if defined (HAVE_LIBPERL) || defined (HAVE_LIBPYTHON)
  "Playlist module directory: " STREAM404_MODULEDIR "\n"
#endif
    );
}

/* Put stream404 in the background, as a daemon */
static void
stream404_setup_daemonize (void)
{
  int stream404pid = fork ();

  if (stream404pid == -1) {
    stream404_log ("ERROR: Cannot fork(), that means no daemon, sorry!");
    return;
  }

  if (stream404pid != 0) {
    /* Update the pidfile (so external applications know what pid
       stream404 is running with. */
    printf ("Into the land of the dreaded daemons we go... (pid: %d)\n", stream404pid);
    stream404_setup_shutdown ();
  }
#ifdef HAVE_SETSID
  setsid ();
#endif

  stream404_log_daemonize ();
  stream404_setup_update_pidfile (getpid());
}

/* Update a file called stream404.pid with the given process id */
static void
stream404_setup_update_pidfile (int stream404pid)
{
  char buf[1024];
  FILE* pidfd;

  if (! stream404_config.base_directory) {
    stream404_log_error ("Base directory is invalid");
    return;
  }

  snprintf (buf, sizeof (buf), "%s/stream404.pid", stream404_config.base_directory);

  pidfd = stream404_util_fopen_for_writing (buf);
	
  if (pidfd) {
    fprintf (pidfd, "%d", stream404pid);
    stream404_util_fclose (pidfd);
  }
}

		
